// News.cpp: implementation of the CNews class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "News.h"

unsigned __stdcall UpdateNews(void* lpParameter);

#define LM_SETLABELTEXT 9600

typedef struct _SETLABELTEXTINFO
{
	string sName, sValue;
} SETLABELTEXTINFO, *LPSETLABELTEXTINFO;
BOOL CALLBACK LabelSearchProc(HWND hwnd, LPARAM lParam);

extern UINT_PTR LastEventIDValue;
extern HWND messageHandler;

void ErrorBox(const char *szMessage);

void SetLabelText(string sLabelName, string sValue);

void DeHTML(string *tempstr, string sNewLine);
string str_replace(string str, const char *a, const char *b);

////////////////////////////////////////////////////////////////////////////
//
// Constructor
// 
CNews::CNews(string sName)
{
	m_sName = sName;
	m_UpdateThread = 0;
	m_bAbort = false;
}

////////////////////////////////////////////////////////////////////////////
//
// Destructor: remove updating thread
// 
CNews::~CNews()
{
	if (m_bActive)
	{
		KillTimer(messageHandler, m_uTimerID);
	}
	m_bAbort = true;
	while (m_UpdateThread) Sleep(50);
}

////////////////////////////////////////////////////////////////////////////
//
// Read settings from .RC. Nothing else
// 
bool CNews::ReadSettings()
{
	char buf[MAX_LINE_LENGTH];

	m_UpdateInterval	= GetRCInt(NAMED_OPTION(m_sName, "UpdateInterval"), 600);

	GetRCLine(NAMED_OPTION(m_sName, "URL"), buf, MAX_LINE_LENGTH, "http://www.shellfront.org/");
	int len = strlen(buf);
	if (len > 0 && buf[len-1] == '\"') buf[len-1] = 0;
	m_sURL = (buf[0] == '\"') ? buf + 1 : buf;

	GetRCString(NAMED_OPTION(m_sName, "InputString"), buf, "{%}", MAX_LINE_LENGTH);
	m_sInputString = str_replace((string)buf, "{%quote}", "\"");

	GetRCString(NAMED_OPTION(m_sName, "OutputString"), buf, "{%1}", MAX_LINE_LENGTH);
	m_sOutputString = str_replace((string)buf, "{%quote}", "\"");

	GetRCString(NAMED_OPTION(m_sName, "NewLine"), buf, "\n", MAX_LINE_LENGTH);
	m_sNewLine = buf;

	GetRCString(NAMED_OPTION(m_sName, "FileNewLine"), buf, "\n", MAX_LINE_LENGTH);
	m_sNewFileLine = buf;

	GetRCLine(NAMED_OPTION(m_sName, "OnChecked"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnChecked = buf;

	GetRCLine(NAMED_OPTION(m_sName, "OnFailure"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnFailure = buf;

	GetRCLine(NAMED_OPTION(m_sName, "OnUpdated"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnUpdated = buf;

	GetRCLine(NAMED_OPTION(m_sName, "DisplayOn"), buf, MAX_LINE_LENGTH, "");
	m_sOutputLabel = buf;

	GetRCLine(NAMED_OPTION(m_sName, "OutputFile"), buf, MAX_LINE_LENGTH, "");
	if (buf[0] && buf[strlen(buf)-1] == '\"') buf[strlen(buf)-1] = 0;
	m_sOutputFile = (buf[0] == '\"') ? buf+1 : buf;

	GetRCLine(NAMED_OPTION(m_sName, "LocalFile"), buf, MAX_LINE_LENGTH, "");
	if (buf[0] && buf[strlen(buf)-1] == '\"') buf[strlen(buf)-1] = 0;
	m_sLocalFile = (buf[0] == '\"') ?  buf+1 : buf;

	m_bActive = GetRCBool(NAMED_OPTION(m_sName, "Enabled"), TRUE) != FALSE;
	return true;
}


////////////////////////////////////////////////////////////////////////////
//
// This function parses HTML file and generates output string
// 
bool CNews::ReadFromFile(const char *szFile, bool bFailIfNotExist)
{
	bool bFailed = false;
	m_bUpdated = false;

	HANDLE f = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (f == INVALID_HANDLE_VALUE)
	{
		if (bFailIfNotExist)
			ErrorBox("Internal error: File successfully downloaded but cannot be open");
		return false;
	}
	DWORD file_length = GetFileSize(f, NULL);
	char *buf = new char[file_length+1];

	// let's start
	try
	{
		DWORD rd;
		if (!ReadFile(f, buf, file_length, &rd, NULL) || rd != file_length)
			throw "Internal error: File successfully opened by cannot be read";

		// first we need to load parameters from the HTML
		buf[file_length] = 0;

		m_params.clear();

		char errmsg[MAX_LINE_LENGTH];
		if (!ReadParameters(buf, file_length, &m_params, errmsg))
		{
			m_params.clear();
			throw errmsg;
		}

		m_bUpdated = false;
		if (m_sOutputFile != "")
		{
			string sOldTextForFile = m_sTextForFile;
			m_sTextForFile = "";

			StringVector params_backup = m_params;

			StringVector::iterator si = m_params.begin(), send = m_params.end();
			for( ; si != send; si++)
			{
				if (stricmp(m_sNewFileLine.c_str(), "\\n") != 0)
				{
					*si = str_replace(*si, "\r\n", m_sNewFileLine.c_str());
					*si = str_replace(*si, "\n", m_sNewFileLine.c_str());
					*si = str_replace(*si, "\r", m_sNewFileLine.c_str());
				}
			}
			if (!GenerateText(&m_params, m_sOutputString, &m_sTextForFile, false))
			{
				m_params.clear();
				throw m_sTextForFile.c_str();
			}
			m_bUpdated = (m_sTextForFile != sOldTextForFile);

			m_params = params_backup;
		}
		if (m_sOutputLabel != "")
		{
			string sOldText = m_sText;
			m_sText = "";
			StringVector::iterator si = m_params.begin(), send = m_params.end();
			for( ; si != send; si++)
			{
				DeHTML(si, m_sNewLine);
				*si = str_replace(*si, ";", "[semicolon]");
			}

			if (!GenerateText(&m_params, m_sOutputString, &m_sText, true))
			{
				m_params.clear();
				throw m_sText.c_str();
			}
			m_bUpdated = m_bUpdated || (m_sText != sOldText);
		}
	} catch (const char *msg)
	{
		// one of our error messages
		char errmsg[MAX_LINE_LENGTH];
		StringCchPrintf(errmsg, MAX_LINE_LENGTH, "The following error occured during web-page parsing:\n%s", msg);
		ErrorBox(errmsg);
		bFailed = true;
	}
	catch (...)
	{
		// unspecified exception
		ErrorBox("Unknown error occured during web-page parsing.");
		bFailed = true;
	}
	// let's stop :)

	delete buf;

	// just close file handle, caller would erase file by itself
	CloseHandle(f);
	return !bFailed;
}


////////////////////////////////////////////////////////////////////////////
//
// Find first escape-sequence in the pattern
// 
char* FindFirstEscapeSequence(const char *szPattern)
{
	char *p1 = strstr(szPattern, "{*}");
	char *p2 = strstr(szPattern, "{%}");
	char *p3 = strstr(szPattern, "{%,");
	if (p2 == NULL || (p2 > p3 && p3 != NULL)) p2 = p3;
	if (p1 == NULL || (p1 > p2 && p2 != NULL)) p1 = p2;
	return p1;
}


////////////////////////////////////////////////////////////////////////////
//
// Read paramteres from HTML using user-specified pattern
// 
bool CNews::ReadParameters(char *szString, int len, StringVector *v, char *errmsg)
{
	char *szBuf = new char[m_sInputString.length()+1];
	StringCchCopy(szBuf, MAX_LINE_LENGTH, m_sInputString.c_str());
	char *szPattern = szBuf;
	int i = 0;	// position of the cursor in the pattern
	int j = 0;	//					and in the HTML file

	while (*szPattern && *szString)
	{
		if (strnicmp(szPattern, "{%}",3) == 0)	// read parameter
		{
			szPattern += 3;
			char *f = FindFirstEscapeSequence(szPattern);
			if (!f)
			{
				f = szPattern + strlen(szPattern);
			} else if (f == szPattern)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "No escape-sequences can follow {%} tag");
				return false;
			}
			bool bFound = false;
			for (char *p = szString; *p; p++)
			{
				if (strnicmp(p, szPattern, f-szPattern) == 0)
				{
					bFound = true;
					break;
				}
			}
			if (!bFound)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "User-specified substring not found");
				return false;
			}
			char *szTmp = new char[p-szString+1];
			strncpy(szTmp, szString, p-szString);
			szTmp[p-szString] = 0;
			v->push_back((string)szTmp);

			delete szTmp;

			szString = p + (f - szPattern);
			szPattern = f;
		} else
		if (strnicmp(szPattern, "{%,",3) == 0)	// read parameter
		{
			szPattern+=3;
			char *err;
			int k = strtod(szPattern, &err);
			if (k<1 || *err != '}')
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "{%,N} tag found but N is incorrect");
				return false;
			}
			if (strlen(szString) < k)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "Parameter must be longer than existing remainder");
				return false;
			}
			char *szTmp = new char[k+1];
			strncpy(szTmp, szString, k);
			szTmp[k] = 0;
			v->push_back((string)szTmp);

			delete szTmp;

			szPattern = strchr(szPattern, '}') + 1;
			szString+=k;
		} else		// skip some symbols
		{
			bool bWithTag = false;
			if (strnicmp(szPattern, "{*}",3) == 0)
			{
				szPattern += 3;
				bWithTag = true;
			}
			char *f = FindFirstEscapeSequence(szPattern);
			if (!f)
			{
				delete szBuf;
				if (bWithTag)
					StringCchCopy(errmsg, MAX_LINE_LENGTH, "{*} is last escape-sequence in the string, so it cannot be applied");
				else
					StringCchCopy(errmsg, MAX_LINE_LENGTH, "Pusycat!");
				return false;
			}
			if (bWithTag && f == szPattern)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "No escape-sequences can follow {*} tag");
				return false;
			}
			bool bFound = false;
			for (char *p = szString; *p; p++)
			{
				if (strnicmp(p, szPattern, f-szPattern) == 0)
				{
					bFound = true;
					break;
				}
			}
			if (!bFound)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "User-specified substring not found");
				return false;
			}
			szString = p + (f - szPattern);
			szPattern = f;
		}
	}
	delete szBuf;
	return true;
}


////////////////////////////////////////////////////////////////////////////
//
// Generate output string using pattern
// 
bool CNews::GenerateText(StringVector *v, string sPattern, string *out, bool bIgnoreNewLine)
{
	char *szBuf = new char[sPattern.length()+1];
	char *szPattern = szBuf;
	StringCchCopy(szBuf, sPattern.length()+1, sPattern.c_str());
	int i = 0;	// position of the cursor in the pattern
	int j = 0;	//					and in the HTML file
	while (*szPattern)
	{
		if (!bIgnoreNewLine && strnicmp(szPattern, "\\\\", 2) == 0)
		{
			szPattern += 2;
			*out += "\\";
		} else
		if (!bIgnoreNewLine && strnicmp(szPattern, "\\n", 2) == 0)
		{
			szPattern +=2;
			*out += "\r\n";
		} else
		if (strnicmp(szPattern, "{%", 2) == 0)
		{
			szPattern += 2;
			char *err;
			int k = strtod(szPattern, &err);

			if (k<1 || *err != '}' || k>v->size())
			{
				delete szBuf;
				*out = "Invalid parameter id";
				return false;
			}
			*out += (*v)[k-1];
			szPattern = strchr(szPattern, '}') + 1;
		} else
		{
			*out += *szPattern;
			szPattern ++;
		}
	}
	delete szBuf;
	return true;
}


////////////////////////////////////////////////////////////////////////////
//
// Enable autoupate
// 
void CNews::Enable()
{
	if (m_bActive) return;
	m_bActive = true;
	LastEventIDValue++;
	m_nIDEvent = LastEventIDValue;
	m_uTimerID = SetTimer(messageHandler, m_nIDEvent, 1000*m_UpdateInterval, NULL);
}

////////////////////////////////////////////////////////////////////////////
//
// Disable autoupdate
// 
void CNews::Disable()
{
	if (!m_bActive) return;
	m_bActive = false;
	KillTimer(messageHandler, m_uTimerID);
}

////////////////////////////////////////////////////////////////////////////
//
// Toggle autoupdate
// 
void CNews::Toggle()
{
	m_bActive ? Disable() : Enable();
}

////////////////////////////////////////////////////////////////////////////
//
// Change input pattern
// 
void CNews::SetInputString(const char *str)
{
	m_sInputString = str_replace((string)str, "{%quote}", "\"");
}

////////////////////////////////////////////////////////////////////////////
//
// Change output pattern
// 
void CNews::SetOutputString(const char *str)
{
	m_sOutputString = str_replace((string)str, "{%quote}", "\"");
	if (m_sLocalFile != "")
	{
		ReadFromFile(m_sLocalFile.c_str(), false);
	}
}

////////////////////////////////////////////////////////////////////////////
//
// Check the news
// 
void CNews::Update()
{
	// if we are not on updating
	if (m_UpdateThread == 0)
	{
		m_bAbort = false;
		_beginthreadex(NULL, 0, UpdateNews, this, 0, &m_UpdateThread);
	}
}

////////////////////////////////////////////////////////////////////////////
//
// Set label text using the pattern specified by user
// 
void CNews::SetLabelText(const char *szLabelName, const char *szPattern)
{
	string sText;
	GenerateText(&m_params, (string)szPattern, &sText, true);
	::SetLabelText((string)szLabelName, sText);
}

////////////////////////////////////////////////////////////////////////////
//
// Replace {%N} with string values
// 
void CNews::ParseCommand(const char *szBang)
{
	string sText;
	string sBang = str_replace((string)szBang, "{%quote}", "\"");
	GenerateText(&m_params, sBang, &sText, true);
	for (int i = 0; i < sText.length(); i++)
	{
		if (sText[i] == '\"') sText[i] = '\'';
	}
	LSExecute(NULL, sText.c_str(), SW_SHOWDEFAULT);
}

////////////////////////////////////////////////////////////////////////////
//
// Set text for the Label, xLabel or xLabelLight window.
// 
void SetLabelText(string sLabelName, string sValue)
{
	static HWND hwndLitestep = GetLitestepWnd();

	SETLABELTEXTINFO slti;
	slti.sName = sLabelName;
	slti.sValue = sValue;
	EnumWindows(LabelSearchProc, (LPARAM)&slti);
}

////////////////////////////////////////////////////////////////////////////
//
// Enumeration
// 
BOOL CALLBACK LabelSearchProc(HWND hwnd, LPARAM lParam)
{
	char buf[MAX_LINE_LENGTH], buf2[MAX_LINE_LENGTH];
	if (GetWindowText(hwnd, buf, MAX_LINE_LENGTH))
	{
		if (buf == ((LPSETLABELTEXTINFO)lParam)->sName )
		{
			if (GetClassName(hwnd, buf2, MAX_LINE_LENGTH))
			{
				if (stricmp(buf2,"xLabel") == 0 || stricmp(buf2, "LabelLS") == 0)
				{
					SendMessage(hwnd, LM_SETLABELTEXT, 0, (LPARAM)(((LPSETLABELTEXTINFO)lParam)->sValue).c_str());
					return FALSE;
				}
			}
		}
	}
	EnumChildWindows(hwnd, LabelSearchProc, lParam);
	return TRUE;
}


bool MyURLDownloadToFile(const char *url, const char *file, bool *bAbort)
{
	DWORD_PTR dwContext = NULL;
	char buf[1024];
	DWORD rd = 0;

	HINTERNET hInternet = InternetOpen(NULL, PRE_CONFIG_INTERNET_ACCESS, NULL, NULL, 0);
	if (*bAbort || !hInternet)
	{
		//ErrorBox("InternetOpen failed");
		return false;
	}

	HINTERNET hFile = InternetOpenUrl(
		hInternet,
		url,
		NULL,
		0,
		INTERNET_FLAG_TRANSFER_BINARY | INTERNET_FLAG_RELOAD | INTERNET_FLAG_NO_CACHE_WRITE,
		dwContext);
	if (*bAbort || !hFile)
	{
		InternetCloseHandle(hInternet);
		return false;
	}

	HANDLE f = CreateFile(
		file,
		GENERIC_WRITE,
		FILE_SHARE_READ,
		NULL,
		CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,
		NULL);
	if (*bAbort || f == INVALID_HANDLE_VALUE)
	{
		//ErrorBox("");
		InternetCloseHandle(hFile);
		InternetCloseHandle(hInternet);
		return false;
	}
	while(InternetReadFile(hFile, buf, 1024, &rd) && rd)
	{
		DWORD wr;
		WriteFile(f, buf, rd, &wr, NULL);
		if (*bAbort || wr != rd)
		{
			CloseHandle(f);
			InternetCloseHandle(hFile);
			InternetCloseHandle(hInternet);
			return false;
		}
	}
	CloseHandle(f);
	InternetCloseHandle(hFile);
	InternetCloseHandle(hInternet);
	return true;
}


////////////////////////////////////////////////////////////////////////////
//
// This thread checks news.
// We cannot download file from the main thread cause it can freeze LS
//
unsigned __stdcall UpdateNews(void* lpParameter)
{
	CNews *news = (CNews*)lpParameter;
	char tempfilename[MAX_LINE_LENGTH] = "";
	try
	{
		if (news->m_sLocalFile != "")
			StringCchCopy(tempfilename, MAX_LINE_LENGTH, news->m_sLocalFile.c_str());
		else
		{
			GetTempPath(MAX_LINE_LENGTH, tempfilename);
			StringCchCat(tempfilename, MAX_LINE_LENGTH, "\\lsnews.html");
		}
		if (!MyURLDownloadToFile(news->m_sURL.c_str(), tempfilename, &(news->m_bAbort)))
			throw "There is no file with specified URL";
		if (!news->ReadFromFile(tempfilename, true)) throw "News file parsing failed";
		if (news->m_sOnChecked != "" &&
			stricmp(news->m_sOnChecked.c_str(), "!none") != 0)
		{
			LSExecute(NULL, news->m_sOnChecked.c_str(), SW_SHOWNORMAL);
			if (news->m_bUpdated &&
				stricmp(news->m_sOnUpdated.c_str(), "!none") != 0 &&
				news->m_sOnUpdated != "" &&
				news->m_sLocalFile != "")
			{
				news->m_bUpdated = false;
				LSExecute(NULL, news->m_sOnUpdated.c_str(), SW_SHOWNORMAL);
			}
		}
		if (news->m_sOutputLabel != "")
			SetLabelText(news->m_sOutputLabel, news->m_sText);
		if (news->m_sOutputFile != "")
		{
			HANDLE f = CreateFile(
				news->m_sOutputFile.c_str(),
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);
			if (f != INVALID_HANDLE_VALUE)
			{
				DWORD wr;
				WriteFile(f, news->m_sTextForFile.c_str(), news->m_sTextForFile.length(), &wr, NULL);
				CloseHandle(f);
			}
		}
	}
	catch(const char *)
	{
		if (stricmp(news->m_sOnFailure.c_str(), "!none") != 0)
		{
			LSExecute(NULL, news->m_sOnFailure.c_str(), SW_SHOWNORMAL);
		}
	}
	catch(...)
	{
		ErrorBox("Internal error: Unknown exception raised");
	}
	news->m_UpdateThread = 0;
	return 1;
}

////////////////////////////////////////////////////////////////////////////
//
// Remove some HTML tags from the string
//
void DeHTML(string *tempstr, string sNewLine)
{
	char *tmp = new char[tempstr->length()+1];
	StringCchCopy(tmp, tempstr->length()+1, tempstr->c_str());
	string str;
	int i=0;
	while (tmp[i])
	{
		if (strnicmp(tmp+i, "&lt;", 4) == 0)
		{
			str += '<';
			i+=4;
		} else if (strnicmp(tmp+i, "&gt;", 4) == 0)
		{
			str += '>';
			i+=4;
		} else if (strnicmp(tmp+i, "&raquo;", 7) == 0)
		{
			str += '';
			i+=7;
		}
		else if (strnicmp(tmp+i, "&amp;", 5) == 0)
		{
			str += '&';
			i+=5;
		} else if (strnicmp(tmp+i, "&quot;", 6) == 0)
		{
			str = '\"';
			i+=6;
		} else if (strnicmp(tmp+i, "<br>", 4) == 0)
		{
			str += sNewLine;
			i+=4;
		} else if (strnicmp(tmp+i, "<br/>", 5) == 0)
		{
			str += sNewLine;
			i+=5;
		} else if (strnicmp(tmp+i, "<br />", 6) == 0)
		{
			str += sNewLine;
			i+=6;
		} else if (strnicmp(tmp+i, "<p>", 3) == 0)
		{
			str += sNewLine;
			i+=3;
		} else if (tmp[i] == '<')
		{
			// skip all unknown tags
			do i++; while (tmp[i]!='>');
			i++;
		} else
		{
			str += tmp[i];
			i++;
		}
	}
	*tempstr = str;
}


////////////////////////////////////////////////////////////////////////////
//
// Replace a with b
//
string str_replace(string str, const char *a, const char *b)
{
	string tmp = "";
	char *s = new char[str.length()+1];
	StringCchCopy(s, str.length()+1, str.c_str());

	char *p = s, *q;

	size_t alen = strlen(a);
	size_t blen = strlen(b);

	if (strncmp(p, a, alen) == 0)
	{
		tmp += b;
		p += alen;
	}

	q = strstr(p, a);
	while (q && *p)
	{
		*q = 0;
		tmp += p;
		tmp += b;
		p = q + alen;
		q = strstr(p, a);
	}

	tmp += p;

	delete s;
	return tmp;
}

